/******************************************************************************
**
**                       W e b C a l c   P r o j e c t
**            Providing an HTML/CSS engine for minimal platforms.
**
**  CASIO fx-9860G SDK Library
**  Copyright (c) 2006 CASIO COMPUTER CO., LTD.
**
**  WebCalc is an HTML/CSS engine for minimal platforms designed to display
**  simple documents using minimal CPU and memory resources, on little or
**  low-depth screens.
**
**  Version		Beta 0.8
**  Author		lephe
**  License		CeCILL
**
******************************************************************************/



/* itoa() : writes an int to a string. Not proper here : should be moved. */
char *itoa(int n, char *str, int base)
{
	int i=1, j=0, x;
	if(n<0) str[j++] = '-', n = -n;
	for(x=n;x;x/=base) j++;
	for(x=n;x;x/=base) str[j-i++] = x%base + '0' + 39*(x%base>9);
	str[j] = 0;
	return str;
}



// Including the project header.
#include "WebCalc.h"

// Global window pointer.
Window *window;

/* Log is disabled for now.
int Debug_handle;
char Debug_buffer[DEBUG_STRSIZE+1];
*/

/**
 *	init()
 *
 *	Inits WebCalc. Loads the fontS, allocates data for the various interpreter modules, and
 *	sets the callback function.
 *	Setting a callback function instead of calling it when main() ends allows to free data
 *	if the user quits the program by launching another application int the main menu, what
 *	can happen at any time when waiting for GetKey() events.
 */

void init(void)
{
	// These symbols are defined by objcopy for static resource linking.
	extern unsigned int binary_resources_font5_fxf_start;
	extern unsigned int binary_resources_font7_fxf_start;
	extern unsigned int binary_resources_font9_fxf_start;
	extern unsigned int binary_resources_fontm_fxf_start;
	struct Font *tex_font = NULL, *tex_sybl = NULL;

	// Displaying a loading message on the screen.
	dclear();
	PrintXY(1,1,(const unsigned char *)"Initializing app...",0);
	Bdisp_PutDisp_DD();

	// Initializing the various modules.
	Window_init();
	allocTracer_init();

	// Activating memory errors.
	memory_seterrors(1);

	// Loading the fonts from the static data.
	tex_font = font_load(&binary_resources_font5_fxf_start, NULL);
	font_load(&binary_resources_font7_fxf_start, NULL);
	font_load(&binary_resources_font9_fxf_start, NULL);
	tex_sybl = font_load(&binary_resources_fontm_fxf_start, NULL);

	// Initializing font and tex.
	font_set_pixel_function(dpixel);
	tex_set_pixel_function(dpixel);
	tex_set_fonts(tex_font, tex_sybl);

/* Log is disabled for now.
	if(memory_exists(DEBUG_LOG)) memory_deletefile(DEBUG_LOG);
	memory_createfile(DEBUG_LOG,DEBUG_SIZ);
	Debug_handle = memory_openfile(DEBUG_LOG,_OPENMODE_WRITE);
*/

	// Saving the settings of the allocation tracer.
	allocTracer_save();
}

/**
 *	quit()
 *
 *	Quits WebCalc. Frees allocated data, unloads font and restores ram.
 */

void quit(void)
{
/* Log is disabled for now.
	memory_closefile(Debug_handle);
*/

	// Freeing all loaded fonts.
	font_free_all();

	// Unloading various modules.
	Window_quit();
}

/**
 *	names()
 *
 *	Finds files corresponding the given wildcard expression and returns an array of
 *	allocated pointers. Argument `filenum` is set if not NULL. If `foldersonly` is non-,
 *	files are ignored and only directories are kept.
 *
 *	@param	wildcard		Wildcard expression for file selection.
 *	@param	filenum			Address of file number variable. Can be NULL.
 *	@param	foldersonly		Specifies whether non-folders entries should be ignored.
 *
 *	@return					Allocated array of matching filenames pointers.
 */

char **names(const char *wildcard, int *filenum, int foldersonly)
{
	// Getting the address as fchar type.
	FONTCHARACTER *address = memory_char2font(wildcard);
	// Using an fchar string to retrieve file names.
	FONTCHARACTER found[30];
	// Using a file information structure.
	FILE_INFO info;
	// The returned pointer.
	char **files;
	// Using a search descriptor and a counter.
	int handle, n = 0;

	// Checking if there is at least one corresponding file.
	if(Bfile_FindFirst(address,&handle,found,&info))
	{
		// If not, closing handles, setting filenum, freeing data and returning.
		Bfile_FindClose(handle);
		if(filenum) *filenum = 0;
		free(address);
		return NULL;
	}
	// Increasing the counter if the found entry matches foldersonly setting.
	if(!foldersonly || info.type == DT_DIRECTORY) n++;

	// Getting the number of valid entries.
	while(!Bfile_FindNext(handle,found,&info))
		if(!foldersonly || info.type == DT_DIRECTORY) n++;

	// Closing the handle.
	Bfile_FindClose(handle);
	// Checking valid entries (needed because of foldersonly setting).
	if(!n)
	{
		// If no valid entry is found, properly return.
		*filenum = 0;
		free(address);
		return NULL;
	}

	// Allocating the main pointer.
	files = (char **)calloc(n,sizeof(char *));

	// Re-searching and copying names this time. Resetting counter.
	n = 0;
	// It is now ensured that a first entry matches.
	Bfile_FindFirst(address,&handle,found,&info);
	// Still checking if it complies with foldersonly setting. Retrieving the name.
	if(!foldersonly || info.type == DT_DIRECTORY) files[n++] = memory_font2char(found);

	// Looping through all the other matching entries.
	while(!Bfile_FindNext(handle,found,&info))
		if(!foldersonly || info.type == DT_DIRECTORY) files[n++] = memory_font2char(found);

	// Setting filenum and closing search descriptor.
	if(filenum) *filenum = n;
	Bfile_FindClose(handle);

	// Freeing the fchar address and returning the main pointer.
	free(address);
	return files;
}

/**
 *	open()
 *
 *	File selection interface. Returns the name of the selected file, in the given folder.
 *	This version is specifically adapted to this program since it filters subfolders and
 *	htm files.
 *
 *	@param	folder	Default folder.
 *
 *	@return			Name of selected file.
 */

const char *open(const char *folder)
{
	// Using an f-key bitmap.
	const unsigned char f1key[24] = { 255, 255, 224, 217, 150, 224, 170, 182, 224, 170,
		178, 224, 170, 148, 224, 169, 182, 192, 219, 150, 128, 255, 255, 0 };
	// Using two address pointers.
	char *address_fold = (char *)malloc(strlen(folder)+3);
	char *address_htms = (char *)malloc(strlen(folder)+7);
	// Using two pointers and two sizes for file lists.
	char **fold, **htms;
	int foldnum, htmsnum;
	// Using a key variable.
	unsigned int key;
	// Using an incremental counter, a list offset and a display offset.
	int i, offset = 0, doffset;
	// Using a temporary pointer and a pointer for the selected file.
	char *tmp, *chosen = NULL;

	// Generating the search strings.
	for(i=0;i<strlen(folder);i++) address_fold[i] = address_htms[i] = folder[i];
	strcpy(address_fold+i,"\\*");
	strcpy(address_htms+i,"\\*.htm");

	// Getting the existing subfolders and htm files.
	fold = names(address_fold, &foldnum, 1);
	htms = names(address_htms, &htmsnum, 0);

	// Checking if data exists.
	if(!fold && !htms)
	{
		// If not, display an error message.
		dclear();
		locate(1,1);
		Print((unsigned char *)"No folder/htm file");
		// Wait for a keyboard event.
		GetKey(&key);

		// Freeing the allocated addresses.
		free(address_fold);
		free(address_htms);
		// Returning NULL means the same as `..`.
		return NULL;
	}

	// This loop is broken when the user selects an entry.
	while(1)
	{
		// Calculating display offset from list offset (cursor position).
		if(foldnum + htmsnum < 8) doffset = 0;
		else if(offset < 4) doffset = 0;
		else if(offset > foldnum+htmsnum-4) doffset = foldnum+htmsnum-7;
		else doffset = offset - 3;

		// Clearing the screen.
		dclear();
		// Displaying the visible filenames.
		for(i=doffset;i<foldnum + htmsnum && i < doffset+7;i++)
		{
			// Locating the cursor, leaving a space for an opening bracket if folder.
			locate(2+(i<foldnum),i-doffset+1);
			// Printing the entry name.
			Print((unsigned char *)(i<foldnum ? fold[i] : htms[i]));
			// If the entry is a folder.
			if(i<foldnum)
			{
				// Printing opening and closing brackets.
				locate(2,i-doffset+1);
				Print((unsigned char *)"[");
				locate(3+strlen(fold[i]),i-doffset+1);
				Print((unsigned char *)"]");
			}
		}
		// Reversing the screen area corresponding to the selected row.
		Bdisp_AreaReverseVRAM(0,(offset-doffset) << 3,127,((offset-doffset)<<3)+7);
		// Displaying the f-key bitmap.
		disp_bitmap(2,56,f1key,19,8);

		// User event loop.
		do
		{
			// Waiting a keyboard event.
			GetKey(&key);
			// Treating up and down keys, respectively.
			if(key == 30018 && offset) { offset--; break; }
			if(key == 30023 && offset < foldnum + htmsnum - 1) { offset++; break; }
		}
		// Exiting on key presses [exe], [exit] and [f1].
		while(key != 30002 && key != 30004 && key != 30009);
		// If key is [exit], preventing the selected entry from not being freed.
		if(key == 30002) { offset=-1; break; }
		// Re-breaking the loop.
		if(key == 30004 || key == 30009) break;
	}

	// Setting the chosen pointer, except if [exit] was pressed.
	if(offset+1)
	{
		// Retrieving the correct pointer.
		tmp = (offset<foldnum ? fold[offset] : htms[offset-foldnum]);
		// Allocating and copying data.
		chosen = strcpy(malloc(strlen(tmp)+1),tmp);
	}

	// Freeing name data.
	for(i=0;i<foldnum;i++) free(fold[i]);
	for(i=0;i<htmsnum;i++) free(htms[i]);
	// Freeing pointer arrays.
	free(fold);
	free(htms);
	// Freeing addresses.
	free(address_fold);
	free(address_htms);

	// Returning the selected name.
	return chosen;
}

static int level(struct DomElement *el)
{
	int level = 0;
	while((el = el->parent)) level++;
	return level;
}

void grand_debug(void)
{
	struct DomElement *el = window->html;
	unsigned key;
	int x, o = 2;

	if(!el)
	{
		dclear();
		locate(1,1);
		Print((unsigned char *)"No html !");
		GetKey(&key);
		return;
	}

	while(1)
	{
		dclear();
		locate(4,1);
		Print((unsigned char *)"document tree");
		Bdisp_AreaReverseVRAM(0,0,127,7);

		x = o;
		Dom_parse(1);
		while((el = Dom_parse(0)))
		{
			locate(level(el)+1,x);
			if(x>1 && x<9) Print((unsigned char *)el->tagName);
			x++;
		}

		GetKey(&key);
		if(key == 30018 && o<2) o++;
		if(key == 30023 && o>-(x-o-8)) o--;
		if(key == 30002) break;
	}
}

/**
 *	main()
 *
 *	Main function in the program, but not the entry point (see initialize in crt0.s).
 *	Handles the main operations such as file loading and main display loop.
 *
 *	@return		Error code.
 */

int main(void)
{
	// Using a key variable.
	unsigned int key;
	// Using a file data pointer and a file size variable.
	char *fileData = NULL;
	int fileSize = 0;
	// Using minimal and current scrolling offsets in the opened document.
	int offset, minOffset;
	// Using a string containing the current directory displayed in file selection.
	char *cwd;

	// Initializing the application.
	init();

	// Main program loop. Theoretically never broken.
	while(1)
	{
		// Clearing the modules.
		Window_clear();
		html_clear();

		// Beginning in the storage memory root.
		cwd = strcpy(malloc(7),"\\\\fls0");
		// The loop is broken when an htm file is selected.
		do
		{
			// Getting an entry in the current folder.
			const char *chosen = open(cwd);

			// If an entry has been selected, chosen is allocated and set.
			if(chosen)
			{
				// Allocating a pointer for concatenation.
				char *tmp = malloc(strlen(cwd)+strlen(chosen)+2);
				// Concatenating the strings with a backslash ('\') between them.
				strcpy(tmp,cwd);
				strcpy(tmp+strlen(tmp),"\\");
				strcpy(tmp+strlen(tmp),chosen);
				// Freeing previously allocated data.
				free(cwd);
				free(chosen);
				// Setting the working directory.
				cwd = tmp;
			}
			// Otherwise, there was nothing to select, or [EXIT] was pressed.
			// Getting to the parent folder if we are not in the root folder.
			else if(strlen(cwd) > 6)
			{
				// Looking for the last folder separator.
				char *tmp = strrchr(cwd,'\\');
				// Cutting the string.
				*tmp = 0;
			}
		}
		while(strcmp(cwd + strlen(cwd) - 4,".htm"));

		// Displaying a loading message.
		dclear();
		PrintXY(1,1,(const unsigned char *)"Loading...",0);
		Bdisp_PutDisp_DD();

		// Loading the file and returning if an error occurs (data is freed by quit()).
		fileData = memory_load(cwd,&fileSize);
		if(!fileData) return 0;

		// Setting the working directory.
		char *tmp = strrchr(cwd,'\\');
		*(tmp+1) = 0;
		Window_setWorkingDirectory(cwd);

		// Displaying a loading message.
		dclear();
		PrintXY(1,1,(const unsigned char *)"Parsing...",0);
		Bdisp_PutDisp_DD();

		// Parsing file data.
		html_parse(fileData, fileData + fileSize);

		// Displaying a loading message.
		dclear();
		PrintXY(1,1,(const unsigned char *)"Applying styles...",0);
		Bdisp_PutDisp_DD();

		// Applying the styles.
		css_parse_all();

		// Displaying a loading message.
		dclear();
		PrintXY(1,1,(const unsigned char *)"Resolving...",0);
		Bdisp_PutDisp_DD();

		// Resizing the window, implicitly calling the DOM solver.
		Window_resize(127, 64);

		// Setting the default offset.
		offset = 0;
		// Computing the minimum offset.
		minOffset = -window->html->style->height + window->height;
		// minOffset might be positive if the document height is less than window->height.
		if(minOffset > 0) minOffset = 0;

		// Setting an optimal key repeat time.
		Bkey_Set_RepeatTime(1, 1);

		// This loop is broken when the document is closed.
		while(1)
		{
			// Displaying the document at the current offset.
			dclear();
			display(offset);
			Bdisp_PutDisp_DD();

			// The loop is broken when a valid event is received.
			while(1)
			{
				// Getting a keyboard event.
				GetKey(&key);
				// Pressing [F6] displays information on RAM usage.
//				if(key==30014) { tools_allocTracer(); break; }
				// Pressing [F5] displays informarion on DOM structure.
				if(key==30013) { grand_debug(); break; }
				// Using top and bottom keys to navigate through the document.
				if(key==30018) { offset+=2; break; }
				if(key==30023) { offset-=2; break; }
				// Using left and right as fast forward.
				if(key==30020) { offset+=8; break; }
				if(key==30021) { offset-=8; break; }
				// Pressing [EXIT] closes the document.
				if(key==30002) break;
			}

			// Correcting offset.
			if(offset > 0) offset = 0;
			else if(offset < minOffset) offset = minOffset;
			// Breaking the loop if an exit signal has been received.
			if(key==30002) break;
		}

		// Restoring the default key repeat time.
		Bkey_Set_RepeatTime_Default();

		// Freeing the file data.
		free(fileData);
	}



	// This point is theoretically never reached.
	return 1;
}
