/**
 *	libfont version 3.1
 *
 *	Written by lephe.
 *	The library consists in this header file and its associated source
 *	file.
 *
 *	The library is released in public domain (CC0).
 *	See [http://creativecommons.org/about/cc0].
 */

#ifndef _LIBFONT_3_1_H
	#define _LIBFONT_3_1_H

// Standard header inclusion.
/*
#include <stdlib.h>
#include <string.h>
*/
#include <fxlib.h>

// Maximum number of fonts.
#ifndef FONT_NUMBER
	#define FONT_NUMBER 10
#endif // FONT_NUMBER

// NULL pointer definition.
#ifndef NULL
	#define NULL ((void *)0)
#endif // NULL

/**
 *	FontGlyph structure
 *
 *	This structure represents a glyph data, as stored in the font file. The
 *	8-bit gap is unused because the data used to be a 4-byte pointer, which
 *	needed to be 4-aligned.
 *	Since the library is now compatible with the static resource handling,
 *	this fourth member was changed to a simple array whose length is
 *	basically undefined but fixed due to width and height members.
 *	The character is only 8-bit for now, which means 256 possible
 *	combinations, that is, ASCII and extended ASCII character range.
 */

struct FontGlyph
{
	unsigned width			:8;
	unsigned height			:8;
	unsigned character		:8;
	unsigned gap			:8;
	char data[];
};

/**
 *	Font structure
 *
 *	This structure handles the meta information on a loaded, such as name
 *	or default height. It gathers the information on all the glyphs in the
 *	font.
 *	A Font structure pointer is needed in almost all the functions in this
 *	library.
 */

struct Font
{
	char name[8];
	unsigned height			:8;
	unsigned italic			:1;
	unsigned weight			:7;
	unsigned glyphNumber	:16;

	struct FontGlyph **glyphs;
};
// A typedef that may be useful to shorten user code.
typedef struct Font Font;

/**
 *	Library functions
 *
 *	The library basically allows to :
 *	--	Load and free font structures from loaded file data.
 *	--	Render text at any position on the screen, with clipping.
 *	--	Calculate the width of a string, as if rendered with any font.
 *	--	Select the most appropriated loaded font for given properties.
 */

// Sets the pixel function.
void font_set_pixel_function(void (*pixel_function)(int x, int y, int color));
// Loads a structure from loaded filedata.
struct Font *font_load(void *address, int *error);
// Frees a Font structure.
void font_free(struct Font *font);
// Frees all the font loaded in the font database.
void font_free_all(void);
// Renders text between pointers str and end.
void font_text_pointer(int x, int y, const char *str, const char *end,
	struct Font *font);
// Renders text between str and the first NUL character.
void font_text(int x, int y, const char *str, struct Font *font);
// Calculates text length using two pointers.
int font_length_pointer(const char *str, const char *end, struct Font *font);
// Calculates text length using a single pointer.
int font_length(const char *str, struct Font *font);
// Returns the length of a single character.
int font_length_character(int character, struct Font *font);
// Returns the font in the database, whose height is nearest the argument.
struct Font *font_nearest(int font_size);

#endif // _LIBFONT_3_1_H
