/****************************************************************/
/*                                                              */
/*                            Memory                            */
/*                                                              */
/*  Description: Fonctions de manipulation de la memoire        */
/*  Auteur:      LePhenixNoir                                   */
/*  Version:     3.0                                            */
/*  Date:        11.06.2014                                     */
/*  Fichier:     memory.c - Code des fonctions                  */
/*                                                              */
/****************************************************************/

#include "fxlib.h"
#include "memory.h"

#ifndef NULL
#define NULL ((void *)0)
#endif // NULL

static int memory_errors;

void memory_seterrors(int e)
{
	memory_errors = (e!=0);
}

void memory_error(const char *from, const char *func, int val)
{
	unsigned int key;
	char info[20] = "<unavailable>";
	if(!memory_errors) return;  

	PopUpWin(6);

	locate(4,2); Print((unsigned char *)"Memory ERROR !!");
	locate(3,4); Print((unsigned char *)"FROM:");
	locate(8,4); Print((unsigned char *)from);
	locate(3,5); Print((unsigned char *)"FUNC:");
	locate(8,5); Print((unsigned char *)func);
	locate(3,6); Print((unsigned char *)"INFO:");
	locate(8,6); Print((unsigned char *)info);
	locate(3,7); Print((unsigned char *)"META:");
	locate(8,7);

	switch(val)
	{
		case  1:  Print((unsigned char *)"NotEnoughRAM"); break;
		case -1:  Print((unsigned char *)"Nonexisting");  break;
		case -5:  Print((unsigned char *)"WrongDevice");  break;
		case -8:  Print((unsigned char *)"AccessDenied"); break;
		case -14: Print((unsigned char *)"ReadOnly");     break;
		case -31: Print((unsigned char *)"DeviceError");  break;
		case -35: Print((unsigned char *)"NotEmpty");     break;
		default:  Print((unsigned char *)"Other");        break;
	}
	GetKey(&key);
}

unsigned short *memory_char2font(const char *adresse)
{
	unsigned short *adr;
	int i;

	adr = (unsigned short *)calloc((strlen(adresse)+1),sizeof(unsigned short));
	if(!adr) { memory_error("char2font()","calloc()",1); return NULL; }
	for(i=0;i<strlen(adresse);i++) *(adr+i) = *(adresse+i);
	return adr;
}

char *memory_font2char(unsigned short *address)
{
	char *str;
	int i, l = 0;

	while(address[l]) l++;
	str = (char *)calloc(l+1, 1);
	if(!str) return NULL;
	for(i=0;address[i];i++) str[i] = address[i];

	return str;
}

int memory_createfile(const char *adresse, int size)
{
	unsigned short *adr = memory_char2font(adresse);
	int i = Bfile_CreateFile(adr,size);
	if(i<0) memory_error("createfile()","CreateFile()",i);

	free(adr);
	return i;
}

int memory_createdir(const char *adresse)
{
	unsigned short *adr = memory_char2font(adresse);
	int i = Bfile_CreateDirectory(adr);
	if(i<0) memory_error("createdir()","CreateDir.()",i);

	free(adr);
	return 1;
}

int memory_openfile(const char *adresse, int mode)
{
	unsigned short *adr = memory_char2font(adresse);
	int i = Bfile_OpenFile(adr,mode);
	if(i<0) memory_error("openfile()","OpenFile()",i);

	free(adr);
	return i;
}

int memory_deletefile(const char *adresse)
{
	unsigned short *adr = memory_char2font(adresse);
	int i = Bfile_DeleteFile(adr);
	if(i<0) memory_error("deletefil.()","DeleteFil.()",i);

	free(adr);
	return i;
}

char **memory_alloc(int l)
{
	char **p = (char **)calloc(l,sizeof(char *));
	int i; for(i=0;i<l;i++) *(p+i) = (char *)calloc(20,1);
	return p;
}

void memory_free(char **p, int l)
{
	int i; for(i=0;i<l;i++) free(*(p+i));
	free(p);
}

int memory_find(char *adresse, char **files, int max)
{
	unsigned short *adr = memory_char2font(adresse);
	unsigned short found[30];
	FILE_INFO fileInfo;
	int searchHandle,i=1,j,x;

	if((x = Bfile_FindFirst(adr,&searchHandle,found,&fileInfo))) return 0;
	for(j=0;j<14 && *(found+j);j++) *(*files+j) = *(found+j);

	while(Bfile_FindNext(searchHandle,found,&fileInfo)==0 && i<max)
	{
		for(j=0;j<14 && *(found+j);j++) *(*(files+i)+j) = *(found+j);
		i++;
	}

	Bfile_FindClose(searchHandle);
	free(adr);
	return i;
}

int memory_exists(char *adresse)
{
	char *file[1];
	int x;

	*file = (char *)malloc(14); **file=0;
	x = memory_find(adresse,file,1);
	free(*file);

	return x!=0;
}

void *memory_load(char *adresse, int *fileSize)
{
	unsigned short *adr = memory_char2font(adresse);
	int handle, x, size;
	void *p;

	if((handle=Bfile_OpenFile(adr,_OPENMODE_READ))<0) { memory_error("load()","OpenFile()",handle); return NULL; }
	size = Bfile_GetFileSize(handle)+1;
	if(fileSize) *fileSize = size;
	p = calloc(size,1);

	if(!p)
	{
		memory_error("load()","malloc()",1);
		Bfile_CloseFile(handle); free(adr); return NULL;
	}
	if((x=Bfile_ReadFile(handle,p,size,0))<0)
	{
		memory_error("load()","ReadFile()",x);
		Bfile_CloseFile(handle); free(adr); return NULL;
	}

	Bfile_CloseFile(handle);
	free(adr);
	return p;
}

int memory_save(char *adresse, void *data, int l)
{
	unsigned short *adr = memory_char2font(adresse);
	int x=0, handle;

	if(memory_exists(adresse)) x = Bfile_DeleteFile(adr);
	if(x<0) { memory_error("save()","DeleteFile()",x); free(adr); return x; }
	x = Bfile_CreateFile(adr,l+1);
	if(x<0) { memory_error("save()","CreateFile()",x); free(adr); return x; }
	handle = Bfile_OpenFile(adr,0x02);
	if(handle<0) { memory_error("save()","OpenFile()",handle); free(adr); return handle; }
	x = memory_writefile(handle,data,l);
	if(x<0) { memory_error("save()","WriteFile()",x); free(adr); return x; }
	memory_closefile(handle);

	free(adr);
	return 0;
}

int memory_user_select(char **files, int n, int exit)
{
	unsigned int key;
	int p=0, offset=0, i;

	while(1)
	{
		Bdisp_AllClr_VRAM();

		for(i=0;i<(n>7?7:n);i++)
			PrintXY(1,9*i+1,(const unsigned char *)*(files+offset+i),0);

		Bdisp_AreaReverseVRAM(0,9*p,100,9*p+8);
		if(offset>0) PrintXY(114,6,(const unsigned char *)"\346\234",0);
		if(offset+7<n) PrintXY(114,51,(const unsigned char *)"\346\235",0);

		while(1)
		{
			GetKey(&key);
			if(key==30002 && exit) return -1;
			if(key==30004) break;
			if(key==30018 && (offset||p))    { if(p>3) p--; else if(offset)     offset--; else p--; break; }
			if(key==30023 && (offset+p+1<n)) { if(p<3) p++; else if(offset+7<n) offset++; else p++; break; }
		}

		if(key==30004) break;
	}

	return offset+p;
}

void *memory_user_autoload(char *prefix, char *selector, int l, int exit, int *filesize)
{
	char **files = memory_alloc(l);
	char *adr = (char *)malloc(strlen(prefix)+strlen(selector)+1);
	void *data;
	int x, handle;

	strcpy(adr,prefix);
	strcpy(adr+strlen(adr),selector);
	x = memory_find(adr,files,l);
	free(adr);
	x = memory_user_select(files,x,exit);
	if(x==-1) return NULL;

	adr = (char *)malloc(strlen(prefix)+strlen(files[x])+1);
	strcpy(adr,prefix);
	strcpy(adr+strlen(adr),files[x]);
	data = memory_load(adr,NULL);

	handle = memory_openfile(adr,_OPENMODE_READ);
	*filesize = memory_filesize(handle);
	memory_closefile(handle);

	free(adr);
	memory_free(files,l);
	return data;
}

